"""
.. module:: mastCrossSearch
    :synopsis: Search MAST Archive for matching observations on a given gPhoton GALEX WCS frame
.. moduleauthor:: Duy Nguyen <dtn5ah@virginia.edu>
"""
# Science Imports
from astropy.io import fits
from astropy.wcs import WCS
from astropy.wcs.utils import pixel_to_skycoord as pix2RaDec
from math import floor
from scipy.spatial import distance

# API Response Imports
import sys
import json
try: # Python 3.x
    from urllib.parse import quote as urlencode
    from urllib.request import urlretrieve
except ImportError:  # Python 2.x
    from urllib import pathname2url as urlencode
    from urllib import urlretrieve
try: # Python 3.x
    import http.client as httplib 
except ImportError:  # Python 2.x
    import httplib 

#TEMPFILE = 'cr_dra_count_coadd.fits'
TEMPFILE = 'Messier060_coadd.fits'


def getFITSCornerCoords(filename):
    fitsImage = fits.open(filename)
    wcs = WCS(fitsImage[0].header)
    rangeX = fitsImage[0].header['NAXIS1']
    rangeY = fitsImage[0].header['NAXIS2']
    corner1 = pix2RaDec(rangeX, 0, wcs)
    corner2 = pix2RaDec(0, 0, wcs)
    corner3 = pix2RaDec(0, rangeY, wcs)
    corner4 = pix2RaDec(rangeX, rangeY, wcs)
    cornerCoords = [corner1, corner2, corner3, corner4]
    return cornerCoords

def getFITSCenterCoords(filename):
    fitsImage = fits.open(filename)
    wcs = WCS(fitsImage[0].header)
    rangeX = fitsImage[0].header['NAXIS1']
    rangeY = fitsImage[0].header['NAXIS2']
    center = pix2RaDec(floor(rangeX / 2.0), floor(rangeY / 2.0), wcs)
    return center

def minCircle(cornerList, center):
    radius = distance.euclidean((cornerList[0].ra.degree, cornerList[0].dec.degree),(center.ra.degree, center.dec.degree))
    return radius

def mastConeQuery(center, radius):
    # Define target server
    server = 'mast.stsci.edu'
    requestsVersion = ".".join(map(str, sys.version_info[:3]))

    # Define request parameters
    requestParams = {'service':'Mast.Caom.Cone',
                     'params':{'ra':center.ra.degree,
                               'dec':center.dec.degree,
                               'radius':radius},
                     'format':'json',
                     'pagesize':2000,
                     'removenullcolumns':True,
                     'timeout':30,
                     'removecache':True}
    JSONquery = urlencode(json.dumps(requestParams))

    # Define HTTP parameters
    httpHeaders = {"Content-type": "application/x-www-form-urlencoded",
                   "Accept": "text/plain",
                   "User-agent":"python-requests/" + requestsVersion}

    # Connect to Server
    connMAST = httplib.HTTPSConnection(server)
    # Invoke MAST Request
    connMAST.request("POST", "/api/v0/invoke", "request="+JSONquery, httpHeaders)
    # Receive MAST Response
    httpResponse = connMAST.getresponse()
    # Return HTTP Response objects
    header = httpResponse.getheaders()
    data = httpResponse.read().decode('utf-8')
    return header,data

def extractGGUIFields(jsonReturn):
    gguiFields = []
    for i, obs in enumerate(jsonReturn['data']):
        mission = obs['obs_collection']
        project = obs['project']
        region = obs['s_region']
        dataType = obs['dataproduct_type']
        gguiFields.append((i, mission, project, dataType, region))
        #print(i, "\t", mission, "\t", project, "\t", dataType, "\n\t\t", region)
    return gguiFields

def appendRegionList(regOutFile, regionParse, paramIgnore):
    shape = regionParse[0]
    ds9Region = regionParse[0].lower() + "("
    #import ipdb; ipdb.set_trace()
    for param in regionParse[paramIgnore:]:
        ds9Region = ds9Region + param + ", "
    ds9Region = ds9Region[:-2] + ")"
    print(ds9Region)

    # DS9 Style Arguments
    mission = obs['obs_collection']
    project = obs['project']
    #import ipdb; ipdb.set_trace()
    if type(project) is type(None): project = "NULL"
    styleArgs = ' ' + '#'
    if project[:4] == "hlsp":
        styleArgs = styleArgs + ' ' + "color=yellow"
    elif mission == "HST" or mission == "HLA":
        styleArgs = styleArgs + ' ' + "color=red"
    elif mission == "KEPLER" or mission == "K2":
        styleArgs = styleArgs + ' ' + "color=green"
    elif mission == "PS1":
        styleArgs = styleArgs + ' ' + "color=blue"
    elif mission == "SWIFT":
        styleArgs = styleArgs + ' ' + "color=cyan"
    elif mission == "GALEX":
        styleArgs = styleArgs + ' ' + "color=magenta"
    else:
        styleArgs = styleArgs + ' ' + "color=green dash=1"
    ds9Region = ds9Region + styleArgs
    print(ds9Region)
    regOutFile.write(ds9Region + "\n")
    

def exportDS9Regions(jsonReturn, fileOutputName):
    gguiFields = []
    regOutFile = open(fileOutputName, "w")

    # DS9 Geometry Syntax
    for i, obs in enumerate(jsonReturn['data']):
        region = obs['s_region']
        regionParse = region.split()
        ds9Region = ''
        paramIgnore = 0
        for param in regionParse:
            try:
                float(param)
                break
            except Exception: paramIgnore+=1; pass
        if regionParse[0] != "CIRCLE" and regionParse[0] != "POLYGON":
            print("Illegal/Unimplemented Shape Detected: ", regionParse[0], ". Skipping Object")
        else:
            #appendRegionList(regOutFile, regionParse, paramIgnore)
            shape = regionParse[0]
            ds9Region = regionParse[0].lower() + "("
            #import ipdb; ipdb.set_trace()
            for param in regionParse[paramIgnore:]:
                ds9Region = ds9Region + param + ", "
            ds9Region = ds9Region[:-2] + ")"
            print(ds9Region)

            # DS9 Style Arguments
            mission = obs['obs_collection']
            project = obs['project']
            #import ipdb; ipdb.set_trace()
            if type(project) is type(None): project = "NULL"
            styleArgs = ' ' + '#'
            if project[:4] == "hlsp":
                styleArgs = styleArgs + ' ' + "color=yellow"
            elif mission == "HST" or mission == "HLA":
                styleArgs = styleArgs + ' ' + "color=red"
            elif mission == "KEPLER" or mission == "K2":
                styleArgs = styleArgs + ' ' + "color=green"
            elif mission == "PS1":
                styleArgs = styleArgs + ' ' + "color=blue"
            elif mission == "SWIFT":
                styleArgs = styleArgs + ' ' + "color=cyan"
            elif mission == "GALEX":
                styleArgs = styleArgs + ' ' + "color=magenta"
            else:
                styleArgs = styleArgs + ' ' + "color=green dash=1"
            ds9Region = ds9Region + styleArgs
            print(ds9Region)
            regOutFile.write(ds9Region + "\n")
    regOutFile.close()
            
# Get Coordinates of Corners
cornerCoords = getFITSCornerCoords(TEMPFILE)
# Get Coordinates of Frame Center
center = getFITSCenterCoords(TEMPFILE)
# Get Radius of smallest enclosing circle of frame
radius = minCircle(cornerCoords, center)

# Perform Mast Cone Query; interpret results
header,data = mastConeQuery(center, radius)
jsonReturn = json.loads(data)
# Extract Relevant information from return
gguiFields = extractGGUIFields(jsonReturn)
#print(gguiFields)
exportDS9Regions(jsonReturn, TEMPFILE + ".reg")



"""
.. module:: qtTabLayouts
    :synopsis: Defines the gGui Target Manager to manage multiple targets
.. moduleauthor:: Duy Nguyen <dnguyen@nrao.edu>
"""

from collections import OrderedDict
from configparser import ConfigParser
from typing import Callable

from PyQt5 import QtWidgets, QtGui
from glue.app.qt.application import GlueApplication
from glue.core.link_helpers import LinkSame
from glue.core.data_factories import load_data

from pkg_resources import resource_filename

class target_manager(QtWidgets.QToolBar):
    """
    Class that handles the loading of gPhoton data and management of multiple
    gGui targets
    """
    def __init__(self, glue_parent: GlueApplication, target_change_callback: Callable[[str], None] = None):
        """Initializes gGui Target Manager
        If provided a dictionary of targets, in outlined gGui YAML structure, it will load those targets into the target manager

        :param glue_parent: The Glue instance that instantiated this object. Target Manager cannot be run outside a Glue context
        :param target_change_callback: Callback function to notify upon primary target change
        """
        super().__init__()
        self._glue_parent = glue_parent
        self._target_catalog = OrderedDict()
        self._primary_name = ""
        self._primary_data = {}
        self._target_change_callbacks = []

        # Initialize GUI Elements
        self.addWidget(QtWidgets.QLabel("gGui Target Manager: "))
        # Add Back Button
        self.addAction(QtGui.QIcon(resource_filename('ggui.icons', 'ArrowBack_transparent.png')), "Previous Target", self.previous_target)
        # Add Combo Box
        self.QComboBox = QtWidgets.QComboBox(self)
        self.QComboBox.currentTextChanged.connect(self.setPrimaryTarget)
        self.addWidget(self.QComboBox)
        # Add Forward Button
        self.addAction(QtGui.QIcon(resource_filename('ggui.icons', 'ArrowForward_transparent.png')), "Next Target", self.next_target)

        # If the initializer wants to know about target changes, register its provided callback
        if target_change_callback:
            self.register_target_change_callback(target_change_callback)

    def register_target_change_callback(self, callback):
        """Registers a callback function to call when primary target changes

        :param callback: Callback function
        """
        self._target_change_callbacks.append(callback)

    def loadTargetDict(self, targDict: dict):
        """Loads a dictionary of targets and associated data product paths into internal cache

        :param targDict: gGui compliant dictionary of targets and paths to associated gPhoton data products
        """
        # Add incoming dictionary to internal cache
        self._target_catalog.update(targDict)
        # Add new items to GUI
        self.QComboBox.addItems(targDict.keys())

    def setPrimaryTarget(self, targName: str):
        """Changes primary target to target specified
        Unloads existing primary target's data (internal cache and parent Glue session), 
        loads the new primary target's data, links their corresponding attributes together,
        and notifies all stakeholders of the new changed primary target

        :param targName: Name of desired new primary target
        """
        # If requested target is not in current cache, throw exception
        if targName not in self._target_catalog.keys():
            raise KeyError("Target Manager does not recognize requested target: " + str(targName))
        # Don't bother doing anything if we're changing to the current target!
        if targName != self._primary_name:
            # If we have data loaded, remove it
            if self._primary_data: 
                def unload_primary_data():
                    for band_data_set in list(self._primary_data.values()):
                        for band_data in band_data_set.values():
                            if isinstance(band_data, list):
                                for data in band_data:
                                    self._glue_parent.data_collection.remove(data)   
                            else:
                                self._glue_parent.data_collection.remove(band_data)                
                unload_primary_data()
            
            # Clear existing target cache
            self._primary_name = None
            self._primary_data.clear()
            
            target_files = self._target_catalog.get(targName)
            # For each gGui Data Type...
            for data_product_type in target_files:
                self._primary_data[data_product_type] = {}

                config = ConfigParser()
                config.read(resource_filename('ggui', 'ggui.conf'))                
                x_att = config.get('Mandatory Fields', data_product_type + "_x", fallback='')
                y_att = config.get('Mandatory Fields', data_product_type + "_y", fallback='')

                # Load every band's data
                for band, band_file in target_files[data_product_type].items():
                    if band_file:
                        self._primary_data[data_product_type][band] = load_data(band_file)
                        
                        # If x_att, y_att provided in conf, test they exist
                        try:
                            if x_att:
                                self._primary_data[data_product_type][band].id[x_att]
                            if y_att:
                                self._primary_data[data_product_type][band].id[y_att]
                        # KeyError means specified attribute doesn't exist in this data. Warn user, and unset attributes, but continue
                        except KeyError as e:
                            parsed_error = e.args[0].split(':')
                            if parsed_error[0]== 'ComponentID not found or not unique':
                                print("WARNING: '" + parsed_error[1].strip() + "' field specified in ggui.conf missing from " + targName + " " + data_product_type + " " + band + ": " + band_file)
                                x_att = ''
                                y_att = ''
                            else:
                                raise
                        # If AttributeError, check if "data" is actually a list of data (multiple data sets per file). Breaks 1-1 correspondence gGui assumes. Warn user plotting and gluing will fail. Skip this data, but import it regardless
                        except AttributeError:
                            if isinstance(self._primary_data[data_product_type][band], list):
                                print("WARNING: " + str(len(self._primary_data[data_product_type][band])) + " datasets imported from " + targName + " " + data_product_type + " band " + band + ". gGui shall import this data, but will be unable to perform automatic actions on it (i.e. gluing, displaying overview, etc.)")
                            else:
                                raise
                            
                        #confirm this is the same object as what was stored in the data collection:
                        self._glue_parent.data_collection.append(self._primary_data[data_product_type][band])
                # If we have multiple bands, glue them together
                try: 
                    if len(self._primary_data[data_product_type].keys()) > 1:
                        attributes_to_glue = {'lightcurve': ['t_mean', 'flux_bgsub'], 
                                            'coadd': ['Right Ascension', 'Declination'], 
                                            'cube': ['Right Ascension', 'Declination', 'World 0']}
                        #for glue_attribute in attributes_to_glue[data_product_type]:
                        for glue_attribute in list(filter(lambda x: x is not '', [x_att, y_att] + config.get('Additional Fields To Glue', data_product_type, fallback='').split(','))):
                            #Can only link two fields at a time. Need to go through all combinations
                            from itertools import permutations
                            for linking_pair in (set(frozenset(t) for t in permutations(self._primary_data[data_product_type].values(),2))):
                                accessor = tuple(linking_pair)
                                self._glue_parent.data_collection.add_link(LinkSame(accessor[0].id[glue_attribute],accessor[1].id[glue_attribute]))
                except TypeError as e:
                    print("Unable to glue " + str(targName) + " " + str(data_product_type) + ": " + str(e))
            
            # Now that all data has been processed properly, officially designate targName as new primary target
            self._primary_name = targName

            # Notify all stakeholders of target change
            for callback in self._target_change_callbacks:
                callback(self._primary_name)

    def getPrimaryData(self) -> dict:
        """Returns currently loaded primary target's data

        :returns: dictionary of the current primary target's data
        """
        return self._primary_data

    def getPrimaryName(self) -> dict:
        """Returns the current primary target's name

        :returns: current primary target's name as string
        """
        return self._primary_name

    def getTargetNames(self) -> list:
        """Returns a list of all cached targets' names

        :returns: list of all cached targets' names
        """
        return self._target_catalog.keys()

    def next_target(self):
        """Advances to next primary target"""
        # Determine the index we're switching to...
        current_target_index = list(self._target_catalog.keys()).index(self._primary_name)
        next_target_index = current_target_index + 1
        # And wrap around to the front if we're currently on the last target
        if next_target_index > int(len(self._target_catalog.keys())) - 1:
            next_target_index = 0
        # Command Target Manager to switch primary targets
        next_target_name = list(self._target_catalog.keys())[next_target_index]
        self.QComboBox.setCurrentText(next_target_name) # QComboBox signal will initiate primary target switching

    def previous_target(self):
        """Advances to previous primary target"""
        # Determine the index we're switching to...
        current_target_index = list(self._target_catalog.keys()).index(self._primary_name)
        next_target_index = current_target_index - 1
        # And wrap around to the back if we're currently on the first target
        if next_target_index < 0:
            next_target_index = int(len(self._target_catalog.keys())) - 1
        # Command Target Manager to switch primary targets
        next_target_name = list(self._target_catalog.keys())[next_target_index]
        self.QComboBox.setCurrentText(next_target_name) # QComboBox signal will initiate primary target switching



"""
.. module:: ggui
    :synopsis: Defines the gGui class and startup behavior
.. moduleauthor:: Duy Nguyen <dnguyen@nrao.edu>
"""

import argparse
import pathlib
import yaml

from glue.core import DataCollection
from glue.app.qt.application import GlueApplication
from glue.config import menubar_plugin
from PyQt5 import QtWidgets

from ggui import qtTabLayouts
from ggui.targetManager import target_manager

class ggui_glue_application(GlueApplication):
    """Primary gGui Application Class
    Integrates gGui framework (target manager, custom tab generation, etc.) into Glue
    """
    
    def __init__(self, data_collection: DataCollection = DataCollection(), target_dict: dict = None):
        """Initializes gGui
        If provided a dictionary of targets, in outlined gGui YAML structure, it will load those targets into the target manager

        :param data_collection: Glue data collection containing Glue data to plot
        :param target_dict: Dict of targets and paths to associated gPhoton data products to load initially
        """
        super().__init__(data_collection)
        # Save a reference to the default tab. We won't need this, but can't delete it until we have multiple tabs
        default_tab = self.current_tab
        
        # Initialize blank overview tab
        def init_overview_tab(self):
            # Initialize blank overview_widget
            self.overview_widget = qtTabLayouts.ggui_overview_tab(session=self.session)
            self.tab_widget.addTab(self.overview_widget, "gGui Overview Tab: No Data Loaded")
            # Set Overview Tab to focus
            self.tab_widget.setCurrentWidget(self.overview_widget)
        init_overview_tab(self)

        # Initialize empty Target Manager
        self.target_manager = target_manager(self, self.primary_target_changed)
        self.addToolBar(self.target_manager)

        if target_dict:
            # Notify user of successful target catalog detection
            print(str(len(target_dict.keys())) + " targets received. Loading " + str(list(target_dict.keys())[0]) + " as default.")

            # Load supplied target catalog into target manager
            # NOTE: Upon first load, Target Manager will automatically update target manager's primary target with the first entry in this dict.
            self.load_targets(target_dict)

        # Delete first default tab
        self.close_tab(self.get_tab_index(default_tab), False)

    def primary_target_changed(self, _):
        """Updates tab data of new primary target
        Indended as signal callback for the target manager to notify gGui of primary target changes
        """
        # Update overview tab with new target's data
        self.overview_widget.load_data(self.session,
                                    self.target_manager.getPrimaryName(),
                                    self.target_manager.getPrimaryData())
        self.tab_widget.setTabText(self.get_tab_index(self.overview_widget), "Overview of " + str(self.target_manager.getPrimaryName()))

    def load_targets(self, target_dict: dict):
        """
        Imports gGui-compliant data (see yaml standard) into target manager

        :param target_dict: Dict of targets and paths to associated gPhoton data products
        """
        self.target_manager.loadTargetDict(target_dict)

    def create_overview_widget(self, target_name: str, target_data: dict):
        """
        Creates an overview tab of gPhoton lightcurve, coadd, and cube data.
        Automatically constructs the tab, adds it to gGui and sets focus to it.

        :param target_name: The name of the target
        :param target_data: The corresponding gPhoton data of the target
        """
        self.overview_widget = qtTabLayouts.ggui_overview_tab(session=self.session, target_name=target_name, target_data=target_data)
        #overview_widget.subWindowActivated.connect(self._update_viewer_in_focus)

        self.tab_widget.addTab(self.overview_widget, "Overview of " + str(target_name))
        # Set Overview Tab to focus
        self.tab_widget.setCurrentWidget(self.overview_widget)

def main(user_arguments: list = None):
    """Entry point/helper function to start ggui

    :param user_arguments: list of arguments, should simulate command line args. Use ['-h'] or ['--help'] for help documentation
    """
    # Initialize argument parser with arguments
    parser = argparse.ArgumentParser(description='gPhoton Graphical User Interface. An analysis package for GALEX gPhoton data products')
    parser.add_argument('--target_list', help='Specify a path to a YAML style list of astronomical targets and associated gPhoton data products')
    parser.add_argument('--yaml_select', action="store_true", help='Spawns a file select dialog to choose a YAML style list of astronomical targets and associated gPhoton data products')
    if user_arguments: 
        args = parser.parse_args(user_arguments)
    else: 
        args = parser.parse_args()
    
    target_data_products = {}

    def validate_targlist_format(target_list: dict,  list_source: str) -> dict:
        empty_targets = []
        for target_name, target_data in target_list.items():
            valid_files = 0
            for data_type, band_data in target_data.items():
                for band, filepathString in band_data.items():
                    if not pathlib.Path(filepathString).is_file():
                        if filepathString:
                            print(filepathString + " does not exist on disk. Ignoring...")
                    else: valid_files += 1
            if not valid_files:
                empty_targets.append(target_name)
        
        for bad_target in empty_targets:
            print(str(bad_target) + " does not have any valid data. Ignoring target...")
            del target_list[bad_target]

        return target_list

    # If the user specified a gGui YAML file, load its targets
    if args.target_list:
        print("File received: " + str(args.target_list))
        target_list_path = pathlib.Path(args.target_list)
        target_data_products.update(validate_targlist_format(yaml.load(open(str(target_list_path), 'r'), Loader=yaml.BaseLoader), str(target_list_path)))
    # If the user requested a file-selector dialog to select a gGui YAML file, display it and load its contents
    if args.yaml_select:
        def prompt_user_for_file(dialogCaption: str, dialogNameFilter: str) -> list:
            """
            Modular QtWidget File-Selection Dialog to prompt user for file import.
            Returns array of filenames selected

            :param dialogCaption: Caption to display along top of file dialog window
            :param dialogNameFilter: Filters file dialog to certain extension
            """
            x = QtWidgets.QApplication([])
            # Set File Dialog Options
            dialog = QtWidgets.QFileDialog(caption=dialogCaption)
            dialog.setFileMode(QtWidgets.QFileDialog.ExistingFiles)
            dialog.setNameFilter(dialogNameFilter)
            # Prompt User for file to import
            dialog.exec_()
            # Get array of File Names
            filenames = dialog.selectedFiles()
            return filenames
        for ggui_yaml_file in prompt_user_for_file("Select GGUI YAML Target List", "gGUI YAML (*.yaml; *.yml)"):
            target_data_products.update(validate_targlist_format(yaml.load(open(ggui_yaml_file, 'r'), Loader=yaml.BaseLoader), ggui_yaml_file))
    # If no targets were recognized, notify the user
    if not target_data_products:
        print("No yaml received. Starting empty gGui session...")
    # Initialize gGui with user-supplied targets, if any
    ggui_app = ggui_glue_application(target_dict=target_data_products)
    # Start gGui
    ggui_app.start()
    
if __name__ == '__main__':
    main()
